#include "korg.hpp"
#include "korgtext.hpp"
#include <string.h>

char* sngname;
FILE* sngf;

void dumpkbdattr(int i, KBDATTR& attr)
{
  printf("KBD%d   Program: %s  Octave: %d\n",
    i+1, place2str(attr.prog), attr.octave+1);
}

void dumpcontrol(int i, TRACKCTRL& trk)
{
  printf("TRACK %d\n", i+1);
  printf("Program:%s  Volume:%d Pan:%s CD SendLevel:%s Channel:%d\n",
    seqplace2str(trk.prog, trk.ctrlfilter), trk.outlevel, pan2str(trk.pan),
    efflev2str(trk.sendlevel, 0),
    trk.channel+1);
  printf("ctrlfilter:%d Transpose:%d Detune:%d\n",
    trk.ctrlfilter, trk.transpose, trk.detune);
  printf("Channel:%d Velocity:%d-%d Keys:%s-",
    trk.channel,
    trk.vel_winbottom, trk.vel_wintop,
    note2str(trk.key_winbottom, -1));
  printf("%s\n", note2str(trk.key_wintop,-1) );
}

int dumpnoteevent(NOTEEVENT& note)
{
    printf("  %5s %-3s  V%03d",
      loc2str(note.loc),
      note2str(note.key, -1),
      note.vel * 2);
    printf("  %s\n", len2str(note.len));
    return 1;
}

int dumpctrlevent(CTRLEVENT& ctrl)
{
  switch(ctrl.cmd)
  {
  case 0x08:
    {
    int v = (((PITCHEVENT*)&ctrl)->v_hi << 7) + ((PITCHEVENT*)&ctrl)->v_lo;

    if (v < 0x2000)
      v |= 0xe000U;
    else
      v &= 0x1fffU;
    printf("  %5s BEND %d\n",
      loc2str(((PITCHEVENT*)&ctrl)->loc),
      v);
    }
    break;
  case 0x10:
    printf("  %5s AFTT %03d\n",
      loc2str(((AFTTEVENT*)&ctrl)->loc),
      ((AFTTEVENT*)&ctrl)->v);
    break;
  case 0x18:
    printf("  %5s PROGRAM %s\n",
      loc2str(((PROGEVENT*)&ctrl)->loc),
      place2str(((PROGEVENT*)&ctrl)->prog));
    break;
  case 0x20:
    if (ctrl.control == 107)
      printf("TEMPO %03d\n", ctrl.value + 30);
    else
      printf("  %5s CTRL %03d %03d\n",
	loc2str(ctrl.loc),
	ctrl.control, ctrl.value);
    break;
  case 0x30:
    printf("BAR %d %s", ((BEATEVENT*)&ctrl)->index+1, tact2str(((BEATEVENT*)&ctrl)->tact));
    if (((BEATEVENT*)&ctrl)->pattern_option != DONT_USE_PATTERN)
    {
      printf("  ===PAT%02d%s===", ((BEATEVENT*)&ctrl)->pattern, ((BEATEVENT*)&ctrl)->pattern_option == PATTERN_START ? "(H)" : "");
    }

    printf("\n");
    break;
  case 0x38:
    printf("END TRACK %d\n", ((ENDEVENT*)&ctrl)->bar);
    return -1;
  default:
    return 0;
  }
  return 1;
}

int dumpevent(EVENT& ev)
{
  switch(ev.general[1] >> 6)
  {
  case 0:
  case 1:
    return dumpctrlevent(ev.ctrl);
  case 2:
    return dumpnoteevent(ev.note);
  case 3:
    return dumpnoteevent(ev.note);
  }
  return 0;
}

void dumpsequence()
{
EVENT ev;

  while (KORGREAD(sngf, ev) && dumpevent(ev))
    ;
}

void dumpsng()
{
SNGCTRL sng[10];
SNGHEAD snghead;
PATCTRL pat;
SNGADR  sngadr;
unsigned *adrtab = (unsigned*)&sngadr;
long adrdiff;

  if (!sngheader(sngf, snghead))
  {
    fprintf(stderr, "%s is not a KORG song file\n", sngname);
    return;
  }
  printf("\nKORG SNG file %s\n\n", sngname);

  for (int i = 0; i <= 9; i++)
  {
    if (!KORGREAD(sngf, sng[i]))
      return;
  }
  for (i = 0; i <= 99; i++)
  {
    if (!KORGREAD(sngf, pat))
      return;
  }
  if (!KORGREAD(sngf, sngadr))
    return;

  for (i = 0; i <= 9; i++)
  {
    printf("sng %d: ", i);
    for (int j = 0; j < 17; j++)
    {
      printf(" %04X", sngadr.sngadr[i][j]);
    }
    printf("\n");
  }
  for (i = 0; i< 100; i++)
  {
    printf(" %04X", sngadr.patadr[i]);
    if (i % 10 == 9)
      printf("\n");
  }
  printf("%04X\n", sngadr.endadr);
  for (i = 0; i <= 9; i++)
  {
    printf("song %d '%.10s' at %ld ",
      i, sng[i].songtitle, ftell(sngf) - sizeof(snghead));

    if ((i < 9 && sngadr.sngadr[i][0] == sngadr.sngadr[i+1][0]) ||
       (i == 9 && sngadr.sngadr[i][0] == sngadr.patadr[0]))
    {
      printf("(empty)\n");
      continue;
    }
    printf("\n");
    for (int t = 0; t < 16; t++)
    if (sngadr.sngadr[i][t] != sngadr.sngadr[i][t+1])
      dumpcontrol(t, sng[i].trackctrl[t]);
  }

  adrdiff = ftell(sngf) - (long(sngadr.sngadr[0][0]) << 4);
  for (adrtab = (unsigned*)&sngadr, i = 0; i < 270; i++, adrtab++)
    if (adrtab[0] != adrtab[1])
    {
   //ofs[i] = ofs[0] + 16 * (v[i] - v[0])
      fseek(sngf, (((long)*adrtab) << 4) + adrdiff, SEEK_SET);
      if (i < 170)
	printf("Song %d track %d at %04X\n", i / 17, (i % 17)+1, ftell(sngf));
      else
	printf("Pattern %d at %04lX\n", i - 170, ftell(sngf));

      dumpsequence();
    }
}

void main(int argc, char**argv)
{
  argc--; argv++;

  if (argc == 0)
    return;

  sngname = *argv++; argc--;

  sngf = fopen(sngname, "rb");
  if (!sngf)
  {
    perror(sngname);
    return;
  }

  dumpsng();

  fclose(sngf);
}
